from __future__ import annotations

from datetime import date, datetime
from typing import TYPE_CHECKING

from clinicedc_constants import (
    ALREADY_REPORTED,
    INVALID_REFERENCE,
    NO,
    NOT_APPLICABLE,
    PRESENT_AT_BASELINE,
    YES,
)
from django import forms
from django.apps import apps as django_apps

from edc_metadata.constants import REQUIRED

from .exceptions import NotEvaluated, ValueBoundryError
from .utils import get_grade_for_value, get_normal_data_or_raise

if TYPE_CHECKING:
    from .models import GradingData, NormalData, ReferenceRangeCollection


class UserFormResponse:
    def __init__(self, utest_id, cleaned_data=None):
        self.utest_id = utest_id
        # ensure each supporting option is provided from the form
        for attr in ["units", "abnormal", "reportable"]:
            if not cleaned_data.get(f"{utest_id}_{attr}"):
                raise forms.ValidationError(
                    {f"{utest_id}_{attr}": "This field is required."}, code=REQUIRED
                )

        self.abnormal = cleaned_data.get(f"{utest_id}_abnormal")
        self.reportable = cleaned_data.get(f"{utest_id}_reportable")
        self.units = cleaned_data.get(f"{utest_id}_units")

    def __repr__(self):
        return f"<UserResponse: {self.utest_id} ({self.units})>"

    def __str__(self):
        return f"{self.utest_id} ({self.units})"


class ReferenceRangeEvaluator:
    def __init__(
        self,
        reference_range_collection_name: str,
        cleaned_data: dict,
        gender: str,
        dob: date,
        report_datetime: datetime,
        age_units: str,
        value_field_suffix: str | None = None,
        **extra_options,
    ):
        if not self.reference_range_collection_model_cls.objects.filter(
            name=reference_range_collection_name
        ).exists():
            raise forms.ValidationError(
                {
                    "__all__": "Invalid reference range collection. "
                    f"Got '{reference_range_collection_name}'"
                },
                code=INVALID_REFERENCE,
            )
        self.reference_range_collection = (
            self.reference_range_collection_model_cls.objects.get(
                name=reference_range_collection_name
            )
        )
        self.cleaned_data = cleaned_data
        self.dob = dob
        self.gender = gender
        self.report_datetime = report_datetime
        self.age_units = age_units
        self.value_field_suffix = value_field_suffix or "_value"
        self.extra_options = extra_options

    def grades(self, utest_id: str) -> list[int]:
        return self.reference_range_collection.grades(utest_id)

    @property
    def reference_range_collection_model_cls(self) -> type[ReferenceRangeCollection]:
        return django_apps.get_model("edc_reportable.referencerangecollection")

    @property
    def normal_data_model_cls(self) -> type[NormalData]:
        return django_apps.get_model("edc_reportable.normaldata")

    def validate_reportable_fields(self):
        """Check normal ranges and grade result values
        for each field mentioned in the reference_range_collection.
        """
        for field, value in self.cleaned_data.items():
            try:
                utest_id, _ = field.split(self.value_field_suffix)
            except ValueError:
                utest_id = field
            if (
                value is not None
                and self.normal_data_model_cls.objects.filter(
                    reference_range_collection=self.reference_range_collection,
                    label=utest_id,
                ).exists()
            ):
                # raise ValidationError if
                self._grade_or_check_normal_range(utest_id, value, field)

    def validate_results_abnormal_field(self):
        """Validate the "results_abnormal" field."""
        self._validate_final_assessment(
            field="results_abnormal",
            responses=[YES],
            suffix="_abnormal",
            word="abnormal",
        )

    def validate_results_reportable_field(self):
        """Validate the "results_reportable" field."""
        self._validate_final_assessment(
            field="results_reportable",
            responses=None,
            suffix="_reportable",
            word="reportable",
        )

    def _grade_or_check_normal_range(self, utest_id: str, value: int | float, field: str):
        """Evaluate a single result value.

        Grading is done first. If the value is not gradeable,
        the value is checked against the normal limits.

        Raise an exception if the user's response is discordant.

        Expected field naming convention:
            * {utest_id}_value
            * {utest_id}_units
            * {utest_id}_abnormal [YES, (NO)]
            * {utest_id}_reportable [(NOT_APPLICABLE), NO, GRADE3, GRADE4]
        """
        # get relevant user form reponses
        user_form_response = UserFormResponse(utest_id, self.cleaned_data)
        opts = dict(
            dob=self.dob,
            gender=self.gender,
            report_datetime=self.report_datetime,
            age_units=self.age_units,
            units=user_form_response.units,
            **self.extra_options,
        )
        grading_data, condition_str = get_grade_for_value(
            reference_range_collection=self.reference_range_collection,
            label=utest_id,
            value=value,
            **opts,
        )
        # is gradeable, user reponse matches grade or has valid opt out
        # response
        if (
            grading_data
            and grading_data.grade
            in self.reference_range_collection.reportable_grades(label=utest_id)
            and str(user_form_response.reportable)
            not in [
                str(grading_data.grade),
                ALREADY_REPORTED,
                PRESENT_AT_BASELINE,
            ]
        ):
            raise forms.ValidationError(
                {
                    field: (
                        f"{utest_id.upper()} is reportable. Got {grading_data.description}. "
                        f"({condition_str})."
                    )
                }
            )
        # user selects grade that does not match grade from evaluator
        if (
            grading_data
            and grading_data.grade
            and str(user_form_response.reportable) in [str(g) for g in self.grades(utest_id)]
            and str(grading_data.grade) != str(user_form_response.reportable)
        ):
            raise forms.ValidationError(
                {
                    field: (
                        f"{utest_id.upper()} grade mismatch. Value given evaluates to grade "
                        f"{grading_data.grade} ({condition_str}). "
                        f"Got grade {user_form_response.reportable}. "
                    )
                }
            )

        # is not gradeable, user reponse is a valid `opt out`.
        if not grading_data and str(user_form_response.reportable) not in [
            NO,
            NOT_APPLICABLE,
        ]:
            raise forms.ValidationError(
                {f"{utest_id}_reportable": "Invalid. Expected 'No' or 'Not applicable'."}
            )
        self._check_normal_range(
            utest_id=utest_id,
            value=value,
            field=field,
            grading_data=grading_data,
            user_form_response=user_form_response,
            opts=opts,
        )

    def _check_normal_range(
        self,
        *,
        utest_id: str,
        value: int | float,
        field: str,
        grading_data: GradingData,
        user_form_response: UserFormResponse,
        opts: dict,
    ):
        try:
            normal_data = get_normal_data_or_raise(
                reference_range_collection=self.reference_range_collection,
                label=utest_id,
                **opts,
            )
        except NotEvaluated as e:
            raise forms.ValidationError({field: str(e)}) from e
        else:
            try:
                is_normal = normal_data.value_in_normal_range_or_raise(
                    value=value,
                    dob=self.dob,
                    report_datetime=self.report_datetime,
                    age_units=self.age_units,
                )
            except ValueBoundryError:
                is_normal = False
            # is not normal, user response does not match
            if not is_normal and user_form_response.abnormal == NO:
                raise forms.ValidationError(
                    {
                        field: (
                            f"{utest_id.upper()} is abnormal. "
                            f"Normal range: {normal_data.description}"
                        )
                    }
                )

            # is not normal and not gradeable, user response does not match
            if is_normal and not grading_data and user_form_response.abnormal == YES:
                raise forms.ValidationError(
                    {f"{utest_id}_abnormal": "Invalid. Result is not abnormal"}
                )

            # illogical user response combination
            if (
                user_form_response.abnormal == YES
                and user_form_response.reportable == NOT_APPLICABLE
            ):
                raise forms.ValidationError(
                    {
                        f"{utest_id}_reportable": (
                            "This field is applicable if result is abnormal"
                        )
                    }
                )

            # illogical user response combination
            if (
                user_form_response.abnormal == NO
                and user_form_response.reportable != NOT_APPLICABLE
            ):
                raise forms.ValidationError(
                    {f"{utest_id}_reportable": "This field is not applicable"}
                )

    def _validate_final_assessment(
        self, *, field: str, suffix: str, word: str, responses: list[str] | None = None
    ):
        """Common code to validate fields `results_abnormal`
        and `results_reportable`.
        """
        responses = responses or [
            str(g) for g in self.reference_range_collection.default_grades()
        ]
        responses = [str(r) for r in responses]
        answers = list(
            {k: v for k, v in self.cleaned_data.items() if k.endswith(suffix)}.values()
        )
        answers = [str(v) for v in answers if v is not None]
        if len(answers) == 0:
            raise forms.ValidationError({"results_abnormal": "No results have been entered."})
        answers_as_bool = [True for v in answers if v in responses]
        if self.cleaned_data.get(field) == NO:
            if any(answers_as_bool):
                are = "is" if len(answers_as_bool) == 1 else "are"
                raise forms.ValidationError(
                    {field: f"{len(answers_as_bool)} of the above results {are} {word}"}
                )
        elif self.cleaned_data.get(field) == YES and not any(answers_as_bool):
            raise forms.ValidationError({field: f"None of the above results are {word}"})



# Generated by Django 5.1.2 on 2024-11-16 16:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0029_remove_historicalrepackrequest_qty_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="historicalrepackrequest",
            name="processed",
        ),
        migrations.RemoveField(
            model_name="repackrequest",
            name="processed",
        ),
        migrations.AddField(
            model_name="historicalrepackrequest",
            name="processed_qty",
            field=models.DecimalField(
                decimal_places=2,
                max_digits=20,
                null=True,
                verbose_name="Number of containers requested",
            ),
        ),
        migrations.AddField(
            model_name="repackrequest",
            name="processed_qty",
            field=models.DecimalField(
                decimal_places=2,
                max_digits=20,
                null=True,
                verbose_name="Number of containers requested",
            ),
        ),
    ]



from clinicedc_constants import NOT_APPLICABLE
from django.contrib import admin
from django.contrib.admin import display
from django.template.loader import render_to_string
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django_audit_fields.admin import audit_fieldset_tuple
from edc_action_item.fieldsets import action_fieldset_tuple
from edc_action_item.modeladmin_mixins import ActionItemModelAdminMixin
from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin
from edc_notification.utils import get_email_contacts
from edc_pdf_reports.admin import PdfButtonModelAdminMixin, print_selected_to_pdf_action

from ..forms import AeInitialForm
from ..models import AeClassification
from ..templatetags.edc_adverse_event_extras import (
    format_ae_description,
    select_description_template,
)
from .list_filters import AeAwarenessListFilter, AeClassificationListFilter
from .modeladmin_mixins import AdverseEventModelAdminMixin

fieldset_part_one = (
    "Part 1: Description",
    {
        "fields": (
            "ae_classification",
            "ae_classification_other",
            "ae_description",
            "ae_awareness_date",
            "ae_start_date",
            "ae_grade",
        )
    },
)

fieldset_part_two = (
    "Part 2: Cause and relationship to study",
    {
        "fields": (
            "ae_study_relation_possibility",
            "study_drug_relation",
            "ae_cause",
            "ae_cause_other",
        )
    },
)

fieldset_part_three = ("Part 3: Treatment", {"fields": ("ae_treatment",)})

fieldset_part_four = (
    "Part 4: SAE / SUSAR",
    {"fields": ("sae", "sae_reason", "susar", "susar_reported")},
)

default_radio_fields = {
    "ae_cause": admin.VERTICAL,
    "ae_classification": admin.VERTICAL,
    "ae_grade": admin.VERTICAL,
    "ae_study_relation_possibility": admin.VERTICAL,
    "study_drug_relation": admin.VERTICAL,
    "sae": admin.VERTICAL,
    "sae_reason": admin.VERTICAL,
    "susar": admin.VERTICAL,
    "susar_reported": admin.VERTICAL,
}


@admin.action(permissions=["view"], description="Print AE Initial Reports as PDF")
def print_to_pdf_action(modeladmin, request, queryset):
    return print_selected_to_pdf_action(modeladmin, request, queryset)


class AeInitialModelAdminMixin(
    PdfButtonModelAdminMixin,
    AdverseEventModelAdminMixin,
    ModelAdminSubjectDashboardMixin,
    ActionItemModelAdminMixin,
):
    form = AeInitialForm

    ordering = ("-ae_awareness_date",)

    actions = (print_to_pdf_action,)

    email_contact = get_email_contacts("ae_reports")
    additional_instructions = format_html(  # nosec B308, B703
        "Complete the initial AE report and forward to the TMG. "
        'Email to <a href="mailto:{}">{}</a>',
        mark_safe(email_contact),  # noqa: S308
        mark_safe(email_contact),  # noqa: S308
    )

    fieldsets = (
        (None, {"fields": ("subject_identifier", "report_datetime")}),
        fieldset_part_one,
        fieldset_part_two,
        fieldset_part_three,
        fieldset_part_four,
        action_fieldset_tuple,
        audit_fieldset_tuple,
    )

    radio_fields = default_radio_fields

    search_fields = (
        "subject_identifier",
        "action_identifier",
        "ae_description",
        "ae_classification__name",
        "ae_classification_other",
    )

    def get_list_display(self, request) -> tuple[str, ...]:
        list_display = super().get_list_display(request)
        custom_fields = (
            "subject_identifier_column",
            "dashboard",
            "pdf_button",
            "description_column",
            "documents_column",
        )
        return custom_fields + tuple(f for f in list_display if f not in custom_fields)

    def get_list_filter(self, request) -> tuple[str, ...]:
        list_filter = super().get_list_filter(request)
        custom_fields = (
            AeAwarenessListFilter,
            "ae_grade",
            AeClassificationListFilter,
            "sae",
            "sae_reason",
            "susar",
            "susar_reported",
        )
        return custom_fields + tuple(f for f in list_filter if f not in custom_fields)

    def get_search_fields(self, request) -> tuple[str, ...]:
        fields = super().get_search_fields(request=request)
        custom_fields = ("ae_description",)
        return custom_fields + tuple(f for f in fields if f not in custom_fields)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "ae_classification":
            kwargs["queryset"] = AeClassification.objects.exclude(name=NOT_APPLICABLE)
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    @display(description="Description", ordering="-ae_awareness_date")
    def description_column(self, obj=None):
        """Returns a formatted comprehensive description of the SAE
        combining multiple fields.
        """
        context = format_ae_description({}, obj, None)
        return render_to_string(select_description_template("aeinitial"), context)

    class Media:
        css = {"all": ("edc_adverse_event/css/extras.css",)}  # noqa: RUF012
        js = ModelAdminSubjectDashboardMixin.Media.js



# Generated by Django 3.2.13 on 2022-09-13 18:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_adherence", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="nonadherencereasons",
            name="plural_name",
            field=models.CharField(max_length=250, null=True, verbose_name="Plural name"),
        ),
    ]



from django.contrib import messages

from edc_dashboard.view_mixins import EdcViewMixin
from edc_lab.labels import AliquotLabel, RequisitionLabel
from edc_lab.models import Aliquot
from edc_label import add_job_results_to_messages

from ...view_mixins import ProcessRequisitionViewMixin
from .action_view import ActionView


class RequisitionView(EdcViewMixin, ProcessRequisitionViewMixin, ActionView):
    post_action_url = "requisition_listboard_url"
    valid_form_actions = ("print_labels", "print_aliquot_labels")
    action_name = "requisition"
    label_cls = RequisitionLabel

    def process_form_action(self, request=None):
        if self.action in self.valid_form_actions:
            if not self.selected_items:
                message = "Nothing to do. No items have been selected."
                messages.warning(request, message)
            elif self.action == "print_labels":
                self._print_labels(request=request)
            elif self.action == "print_aliquot_labels":
                self._print_aliquot_labels(request=request)

    def _print_labels(self, request=None):
        job_results = []
        pks = [
            (obj._meta.label_lower, obj.pk)
            for obj in self.get_requisitions(pk__in=self.selected_items)
        ]
        if pks:
            job_results.append(self.print_labels(pks=pks, request=request))
        if job_results:
            add_job_results_to_messages(request, job_results)

    def _print_aliquot_labels(self, request=None):
        self.label_cls = AliquotLabel
        job_results = []
        for requisition in self.processed_requisitions:
            aliquots = Aliquot.objects.filter(
                requisition_identifier=requisition.requisition_identifier
            ).order_by("count")
            if aliquots:
                pks = [obj.pk for obj in aliquots if obj.is_primary]
                if pks:
                    job_results.append(self.print_labels(pks=pks, request=request))
                pks = [obj.pk for obj in aliquots if not obj.is_primary]
                if pks:
                    job_results.append(self.print_labels(pks=pks, request=request))
        for requisition in self.unprocessed_requisitions:
            messages.error(
                self.request,
                "Unable to print labels. Requisition has not been "
                f"processed. Got {requisition.requisition_identifier}",
            )
        if job_results:
            add_job_results_to_messages(request, job_results)



from .constants import FILL_ACROSS

model_data = {
    "edc_lab.boxtype": [
        {
            "name": "9 x 9",
            "across": 9,
            "down": 9,
            "total": 81,
            "fill_order": FILL_ACROSS,
        }
    ]
}



from clinicedc_constants import NOT_APPLICABLE
from django.conf import settings
from django.db import models
from django.db.models import UniqueConstraint
from django.utils import timezone

from edc_consent.model_mixins import RequiresConsentFieldsModelMixin
from edc_constants.choices import YES_NO
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_metadata.model_mixins.updates import UpdatesRequisitionMetadataModelMixin
from edc_model.models import HistoricalRecords, InitialsField, OtherCharField
from edc_model.validators import datetime_not_future
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_protocol.validators import datetime_not_before_study_start
from edc_search.model_mixins import SearchSlugModelMixin
from edc_sites.model_mixins import SiteModelMixin
from edc_visit_tracking.managers import CrfCurrentSiteManager
from edc_visit_tracking.model_mixins import (
    PreviousVisitModelMixin,
    VisitTrackingRequisitionModelMixin,
)

from ...choices import ITEM_TYPE, REASON_NOT_DRAWN
from ...managers import RequisitionManager
from ..panel_model_mixin import PanelModelMixin
from .requisition_identifier_mixin import RequisitionIdentifierMixin
from .requisition_status_mixin import RequisitionStatusMixin
from .requisition_verify_model_mixin import RequisitionVerifyModelMixin


class RequisitionModelMixin(
    SiteModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    VisitTrackingRequisitionModelMixin,
    PanelModelMixin,
    PreviousVisitModelMixin,
    RequiresConsentFieldsModelMixin,
    RequisitionIdentifierMixin,
    RequisitionStatusMixin,
    RequisitionVerifyModelMixin,
    SearchSlugModelMixin,
    UpdatesRequisitionMetadataModelMixin,
    models.Model,
):
    requisition_datetime = models.DateTimeField(
        validators=[datetime_not_before_study_start, datetime_not_future],
        default=timezone.now,
        verbose_name="Requisition Date",
    )

    drawn_datetime = models.DateTimeField(
        verbose_name="Date / Time Specimen Drawn",
        validators=[datetime_not_before_study_start, datetime_not_future],
        null=True,
        blank=True,
        help_text="If not drawn, leave blank.",
    )

    is_drawn = models.CharField(
        verbose_name="Was a specimen drawn?",
        max_length=3,
        choices=YES_NO,
        help_text="If No, provide a reason below",
    )

    reason_not_drawn = models.CharField(
        verbose_name="If not drawn, please explain",
        max_length=25,
        default=NOT_APPLICABLE,
        choices=REASON_NOT_DRAWN,
    )

    reason_not_drawn_other = OtherCharField()

    protocol_number = models.CharField(max_length=10, default="", editable=False)

    clinician_initials = InitialsField(default="", blank=True)

    specimen_type = models.CharField(
        verbose_name="Specimen type", max_length=25, default="", blank=True
    )

    item_type = models.CharField(
        verbose_name="Item collection type",
        max_length=25,
        choices=ITEM_TYPE,
        default=NOT_APPLICABLE,
        help_text="",
    )

    item_count = models.IntegerField(
        verbose_name="Number of items",
        null=True,
        blank=True,
        help_text=(
            "Number of tubes, samples, cards, etc being sent for this test/order only. "
            "Determines number of labels to print"
        ),
    )

    estimated_volume = models.DecimalField(
        verbose_name="Estimated volume in mL",
        max_digits=7,
        decimal_places=2,
        null=True,
        blank=True,
        help_text=(
            "If applicable, estimated volume of sample for this test/order. "
            'This is the total volume if number of "tubes" above is greater than 1'
        ),
    )

    comments = models.TextField(max_length=25, blank=True, default="")

    objects = RequisitionManager()

    on_site = CrfCurrentSiteManager()

    history = HistoricalRecords(inherit=True)

    def __str__(self):
        return f"{self.panel_object.verbose_name}: {self.requisition_identifier}"

    def save(self, *args, **kwargs):
        if not self.id:
            self.protocol_number = ResearchProtocolConfig().protocol_number
        self.subject_identifier = self.related_visit.subject_identifier
        self.specimen_type = self.panel_object.aliquot_type.alpha_code
        self.report_datetime = self.requisition_datetime
        super().save(*args, **kwargs)

    def natural_key(self):
        return (self.requisition_identifier,)

    natural_key.dependencies = (settings.SUBJECT_VISIT_MODEL, "sites.Site")

    def get_search_slug_fields(self):
        fields = super().get_search_slug_fields()
        return (
            *fields,
            "subject_identifier",
            "requisition_identifier",
            "human_readable_identifier",
            "identifier_prefix",
        )

    class Meta(NonUniqueSubjectIdentifierFieldMixin.Meta):
        abstract = True
        constraints = (
            UniqueConstraint(
                fields=["panel", "subject_visit"],
                name="%(app_label)s_%(class)s_panel_uniq",
            ),
        )
        indexes = (
            *NonUniqueSubjectIdentifierFieldMixin.Meta.indexes,
            models.Index(fields=["subject_visit", "site", "panel"]),
            models.Index(fields=["subject_visit", "panel"]),
            models.Index(fields=["subject_visit", "report_datetime"]),
        )



# Generated by Django 5.2.10 on 2026-01-24 19:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0114_historicalorderitem_container_qty_and_more"),
    ]

    operations = [
        migrations.RenameField(
            model_name="historicalorderitem",
            old_name="container_qty",
            new_name="container_unit_qty",
        ),
        migrations.RenameField(
            model_name="historicalreceiveitem",
            old_name="container_qty",
            new_name="container_unit_qty",
        ),
        migrations.RenameField(
            model_name="historicalstock",
            old_name="container_qty",
            new_name="container_unit_qty",
        ),
        migrations.RenameField(
            model_name="orderitem",
            old_name="container_qty",
            new_name="container_unit_qty",
        ),
        migrations.RenameField(
            model_name="receiveitem",
            old_name="container_qty",
            new_name="container_unit_qty",
        ),
        migrations.RenameField(
            model_name="stock",
            old_name="container_qty",
            new_name="container_unit_qty",
        ),
        migrations.AddField(
            model_name="container",
            name="qty_may_be_adjusted_from_default",
            field=models.BooleanField(
                default=False,
                verbose_name="Qty may be adjusted from the default on order / receive",
            ),
        ),
        migrations.AddField(
            model_name="historicalcontainer",
            name="qty_may_be_adjusted_from_default",
            field=models.BooleanField(
                default=False,
                verbose_name="Qty may be adjusted from the default on order / receive",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorderitem",
            name="qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Number of containers",
                max_digits=20,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalorderitem",
            name="unit_qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="unit qty ordered less unit qty received. Updated in a signal",
                max_digits=20,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="orderitem",
            name="qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Number of containers",
                max_digits=20,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="orderitem",
            name="unit_qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="unit qty ordered less unit qty received. Updated in a signal",
                max_digits=20,
                null=True,
            ),
        ),
    ]



from edc_auth.site_auths import site_auths
from edc_auth.utils import remove_default_model_permissions_from_edc_permissions
from edc_lab.auth_objects import LAB, LAB_VIEW
from edc_lab_dashboard.auth_objects import (
    lab_dashboard_codenames,
    lab_dashboard_tuples,
    lab_navbar_codenames,
    lab_navbar_tuples,
)


def update_site_auths():
    site_auths.add_post_update_func(
        "edc_lab_dashboard", remove_default_model_permissions_from_edc_permissions
    )

    site_auths.add_custom_permissions_tuples(
        model="edc_lab_dashboard.edcpermissions", codename_tuples=lab_navbar_tuples
    )
    site_auths.add_custom_permissions_tuples(
        model="edc_lab_dashboard.edcpermissions", codename_tuples=lab_dashboard_tuples
    )

    site_auths.update_group(*lab_dashboard_codenames, *lab_navbar_codenames, name=LAB)
    site_auths.update_group(*lab_dashboard_codenames, *lab_navbar_codenames, name=LAB_VIEW)


update_site_auths()



from django.contrib.admin.decorators import register

from edc_action_item.admin import ActionItemAdmin as BaseActionItemAdmin

from ..admin_site import edc_data_manager_admin
from ..models import DataManagerActionItem


@register(DataManagerActionItem, site=edc_data_manager_admin)
class DataManagerActionItemAdmin(BaseActionItemAdmin):
    pass



from clinicedc_constants import MICRO_IU_MILLILITER
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models

from edc_constants.choices import YES_NO
from edc_glucose.model_mixin_factories import fasting_model_mixin_factory
from edc_reportable.units import MICRO_IU_MILLILITER_DISPLAY

from ..model_mixin_factories import reportable_result_model_mixin_factory


class InsulinModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="ins",
        verbose_name="Insulin",
        units_choices=((MICRO_IU_MILLILITER, MICRO_IU_MILLILITER_DISPLAY),),
        validators=[MinValueValidator(0.0), MaxValueValidator(999.0)],
    ),
    fasting_model_mixin_factory(
        None,
        field_options={"fasting": dict(choices=YES_NO, default=None)},
        fasting=models.CharField(
            verbose_name="Has the participant fasted?",
            max_length=15,
            choices=YES_NO,
            null=True,
            blank=False,
            help_text="As reported by patient",
        ),
    ),
    models.Model,
):
    class Meta:
        abstract = True



from uuid import uuid4

from faker import Faker
from model_bakery.recipe import Recipe

from .models import DataQuery, QueryRule

fake = Faker()

queryrule = Recipe(QueryRule, title=str(uuid4()))

dataquery = Recipe(
    DataQuery,
    action_identifier=None,
    parent_action_item=None,
    related_action_item=None,
    title=str(uuid4()),
)



# Generated by Django 5.2.1 on 2025-06-02 18:20

import _socket
import django.core.validators
import django.db.models.deletion
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_audit_fields.models.audit_model_mixin
import django_revision.revision_field
from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("edc_reportable", "0002_alter_gradingdata_age_upper_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="referencerangecollection",
            name="grade1",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="referencerangecollection",
            name="grade2",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="referencerangecollection",
            name="grade3",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="referencerangecollection",
            name="grade4",
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="age_phrase",
            field=models.CharField(help_text="calculated in save()", max_length=25, null=True),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="gender",
            field=models.CharField(
                max_length=1,
                validators=[django.core.validators.RegexValidator("[MF]{1}")],
            ),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="phrase",
            field=models.CharField(
                help_text="calculated by the formula instance",
                max_length=50,
                null=True,
                verbose_name="Value phrase",
            ),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="age_phrase",
            field=models.CharField(help_text="calculated in save()", max_length=25, null=True),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="gender",
            field=models.CharField(
                max_length=1,
                validators=[django.core.validators.RegexValidator("[MF]{1}")],
            ),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="grade",
            field=models.IntegerField(
                null=True,
                validators=[
                    django.core.validators.MinValueValidator(1),
                    django.core.validators.MaxValueValidator(5),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="phrase",
            field=models.CharField(
                help_text="calculated by the formula instance",
                max_length=50,
                null=True,
                verbose_name="Value phrase",
            ),
        ),
        migrations.CreateModel(
            name="GradingException",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("label", models.CharField(max_length=50)),
                ("grade1", models.BooleanField(default=False)),
                ("grade2", models.BooleanField(default=False)),
                ("grade3", models.BooleanField(default=False)),
                ("grade4", models.BooleanField(default=False)),
                (
                    "reference_range_collection",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_reportable.referencerangecollection",
                    ),
                ),
            ],
            options={
                "verbose_name": "Grading Exception",
                "verbose_name_plural": "Grading Exceptions",
                "abstract": False,
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "default_manager_name": "objects",
                "indexes": [
                    models.Index(
                        fields=["modified", "created"],
                        name="edc_reporta_modifie_a16896_idx",
                    ),
                    models.Index(
                        fields=["user_modified", "user_created"],
                        name="edc_reporta_user_mo_8ec7a5_idx",
                    ),
                ],
            },
        ),
    ]



from edc_notification.decorators import register
from edc_notification.notification import GradedEventNotification, ModelNotification

from .utils import get_adverse_event_app_label


@register()
class AeInitialG3EventNotification(GradedEventNotification):
    name = "g3_aeinitial"
    display_name = "Grade 3 initial event reported"
    grade = 3
    model = f"{get_adverse_event_app_label()}.aeinitial"


@register()
class AeFollowupG3EventNotification(GradedEventNotification):
    name = "g3_aefollowup"
    display_name = "Grade 3 followup event reported"
    grade = 3
    model = f"{get_adverse_event_app_label()}.aefollowup"


@register()
class AeInitialG4EventNotification(GradedEventNotification):
    name = "g4_aeinitial"
    display_name = "Grade 4 initial event reported"
    grade = 4
    model = f"{get_adverse_event_app_label()}.aeinitial"


@register()
class AeFollowupG4EventNotification(GradedEventNotification):
    name = "g4_aefollowup"
    display_name = "Grade 4 followup event reported"
    grade = 4
    model = f"{get_adverse_event_app_label()}.aefollowup"


@register()
class DeathNotification(ModelNotification):
    name = "death"
    display_name = "a death has been reported"
    model = f"{get_adverse_event_app_label()}.deathreport"



from django.core.exceptions import DisallowedHost


class IgnoreSpecificIPDisallowedHost:
    def __init__(self, ip_to_ignore):
        self.ip_to_ignore = ip_to_ignore

    def filter(self, record):
        if record.exc_info:
            _, exc_value, _ = record.exc_info
            if isinstance(exc_value, DisallowedHost):
                request = getattr(record, "request", None)
                if request and request.META.get("REMOTE_ADDR") == self.ip_to_ignore:
                    return False  # Suppress this log
        return True  # Allow all other logs



from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_sites.managers import CurrentSiteManager
from edc_sites.model_mixins import SiteModelMixin

from ..managers import SubjectRefusalManager
from ..model_mixins import SubjectRefusalModelMixin


class SubjectRefusal(SubjectRefusalModelMixin, SiteModelMixin, BaseUuidModel):
    objects = SubjectRefusalManager()

    on_site = CurrentSiteManager()

    history = HistoricalRecords()

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Subject Refusal"
        verbose_name_plural = "Subject Refusals"



# Generated by Django 5.2.3 on 2025-07-01 03:17

from django.db import migrations, models

import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0076_historicalstockadjustment_stockadjustment"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalstockadjustment",
            name="adjustment_datetime",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AddField(
            model_name="stockadjustment",
            name="adjustment_datetime",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]



from typing import Protocol, TypeVar
from uuid import UUID

from django.db import models
from django_audit_fields.stubs import AuditModelStub


class ModelMetaStub(Protocol):
    verbose_name: str
    verbose_name_plural: str
    label_lower: str
    object_name: str
    app_label: str
    model_name: str

    def get_fields(self) -> list: ...


class BaseUuidModelStub(AuditModelStub, Protocol):
    id: UUID
    admin_url_name: str
    admin_site_name: str

    def get_absolute_url(self) -> str: ...

    objects: models.Manager
    _meta: ModelMetaStub

    ...


TBaseUuidModelStub = TypeVar("TBaseUuidModelStub", bound="BaseUuidModelStub")


class BaseUuidHistoryModelStub(AuditModelStub, Protocol):
    id: UUID
    admin_url_name: str
    admin_site_name: str

    def get_absolute_url(self) -> str: ...

    objects: models.Manager
    history: models.Manager
    _meta: ModelMetaStub

    ...



# Generated by Django 5.2.10 on 2026-01-24 22:03
from decimal import Decimal

from django.db import migrations
from django.db import transaction
from tqdm import tqdm

from edc_pharmacy.models import Container, OrderItem, ReceiveItem, Stock
from django.db.models import F


def update_pharmacy(apps, schema_editor):

    # update new container_unit_qty from container default qty
    # before dropping container default qty
    with transaction.atomic():
        for obj in OrderItem.objects.all():
            obj.container_unit_qty = obj.container.unit_qty_default
            obj.save_base(update_fields=["container_unit_qty"])

    # update new container_unit_qty from container default qty
    # before dropping container default qty
    with transaction.atomic():
        for obj in ReceiveItem.objects.all():
            obj.container_unit_qty = obj.container.unit_qty_default
            obj.save_base(update_fields=["container_unit_qty"])

    # update mising stock unit_qty_in from container default qty
    # before dropping container default qty
    with transaction.atomic():
        qs = Stock.objects.exclude(unit_qty_in=F("container__unit_qty_default"))
        for obj in tqdm(qs, total=qs.count()):
            obj.unit_qty_in = obj.container.unit_qty_default
            obj.save_base(update_fields=["unit_qty_in"])

    # consolidate 'bucket' containers
    with transaction.atomic():
        try:
            container = Container.objects.get(name="BUCKET 16000")
        except Container.DoesNotExist:
            pass
        else:
            container.name = "bucket_of_tablets"
            container.display_name = "Bucket of tablets (Max. 30000)"
            container.save()
            container.refresh_from_db()

            for name in [
                "bucket_11978",
                "bucket_14664",
                "bucket_16550",
                "bucket_9195",
                "BUCKET-16000",
                "bottle30k"
            ]:
                OrderItem.objects.filter(
                    container__name=name,
                    container_unit_qty=F("container__unit_qty_default")
                ).update(container=container)
                ReceiveItem.objects.filter(
                    container__name=name,
                    container_unit_qty=F("container__unit_qty_default")
                ).update(container=container)
                Stock.objects.filter(
                    container__name=name,
                    unit_qty_in=F("container__unit_qty_default"),
                    container_unit_qty=F("container__unit_qty_default")
                ).update(container=container)

        Stock.objects.filter(
            container_unit_qty__isnull=True, container__name="bottle128"
            ).update(container_unit_qty=Decimal("128.0"))
        Stock.objects.filter(container_unit_qty__isnull=True
            ).update(container_unit_qty=F("unit_qty_in"))


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0119_alter_container_max_items_per_subject_and_more"),
    ]

    operations = [migrations.RunPython(update_pharmacy)]



# Generated by Django 2.0 on 2018-01-03 11:51

import django_crypto_fields.fields.encrypted_char_field
import django_crypto_fields.fields.encrypted_text_field
from django.db import migrations, models

import edc_model.validators.phone


class Migration(migrations.Migration):
    dependencies = [("edc_locator", "0002_auto_20180103_1322")]

    operations = [
        migrations.AddField(
            model_name="historicalsubjectlocator",
            name="may_call_work",
            field=models.CharField(
                choices=[("Yes", "Yes"), ("No", "No")],
                default="Yes",
                max_length=25,
                verbose_name="Has the participant given permission to contacted <b>at work</b> by telephone or cell by study staff for follow-up purposes during the study?",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="subjectlocator",
            name="may_call_work",
            field=models.CharField(
                choices=[("Yes", "Yes"), ("No", "No")],
                default="Yes",
                max_length=25,
                verbose_name="Has the participant given permission to contacted <b>at work</b> by telephone or cell by study staff for follow-up purposes during the study?",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="subject_work_phone",
            field=django_crypto_fields.fields.encrypted_char_field.EncryptedCharField(
                blank=True,
                help_text=" (Encryption: RSA local)",
                max_length=71,
                null=True,
                verbose_name="Work contact number",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="subject_work_place",
            field=django_crypto_fields.fields.encrypted_text_field.EncryptedTextField(
                blank=True,
                help_text=" (Encryption: AES local)",
                max_length=250,
                null=True,
                validators=[edc_model.validators.phone.telephone_number],
                verbose_name="Name and location of work place",
            ),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="subject_work_phone",
            field=django_crypto_fields.fields.encrypted_char_field.EncryptedCharField(
                blank=True,
                help_text=" (Encryption: RSA local)",
                max_length=71,
                null=True,
                verbose_name="Work contact number",
            ),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="subject_work_place",
            field=django_crypto_fields.fields.encrypted_text_field.EncryptedTextField(
                blank=True,
                help_text=" (Encryption: AES local)",
                max_length=250,
                null=True,
                validators=[edc_model.validators.phone.telephone_number],
                verbose_name="Name and location of work place",
            ),
        ),
    ]



from dataclasses import dataclass, field
from string import Template


class CrfSubqueryError(Exception):
    pass


@dataclass(kw_only=True)
class CrfSubquery:
    label: str = None
    label_lower: str = None
    dbtable: str = None
    fld_name: str | None = None
    subjectvisit_dbtable: str | None = None
    where: str | None = None
    list_tables: list[tuple[str, str, str]] | None = field(default_factory=list)
    template: Template = field(
        init=False,
        default=Template(
            "select v.subject_identifier, crf.id as original_id, crf.subject_visit_id, "
            "crf.report_datetime, crf.site_id, v.visit_code, "
            "v.visit_code_sequence, v.schedule_name, crf.modified, "
            "'${label_lower}' as label_lower, "
            "'${label}' as label, count(*) as records "
            "from ${dbtable} as crf "
            "left join ${subjectvisit_dbtable} as v on v.id=crf.subject_visit_id "
            "${left_joins} "
            "where ${where} "
            "group by v.subject_identifier, crf.subject_visit_id, crf.report_datetime, "
            "crf.site_id, v.visit_code, v.visit_code_sequence, v.schedule_name, crf.modified"
        ),
    )

    def __post_init__(self):
        # default where statement if not provided and have fld_name.
        if self.where is None and self.fld_name:
            self.where = f"crf.{self.fld_name} is null"
        if not self.label_lower:
            raise CrfSubqueryError("label_lower is required")
        if not self.subjectvisit_dbtable:
            self.subjectvisit_dbtable = f"{self.label_lower.split('.')[0]}_subjectvisit"

    @property
    def left_joins(self) -> str:
        """Add list tbls to access list cols by 'name' instead of 'id'"""
        left_join = []
        for opts in self.list_tables or []:
            list_field, list_dbtable, alias = opts
            left_join.append(
                f"left join {list_dbtable} as {alias} on crf.{list_field}={alias}.id"
            )
        return " ".join(left_join)

    @property
    def sql(self):
        opts = {k: v for k, v in self.__dict__.items() if v is not None}
        opts.update(left_joins=self.left_joins)
        try:
            sql = self.template.substitute(**opts).replace(";", "")
        except KeyError as e:
            raise CrfSubqueryError(e) from e
        return sql



from django.db import models
from django.utils.text import slugify

from edc_model.models import BaseUuidModel


class ListModelManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)


class BaseListModelMixin(models.Model):
    # FIXME: this should be a short string, e.g. 15-25 chars!
    name = models.CharField(
        verbose_name="Stored value",
        max_length=250,
        unique=True,
        help_text="This is the stored value, required",
    )

    plural_name = models.CharField(
        verbose_name="Plural name",
        max_length=250,
        default="",
    )

    display_name = models.CharField(
        verbose_name="Name",
        max_length=250,
        unique=True,
        help_text="(suggest 40 characters max.)",
    )

    display_index = models.IntegerField(
        verbose_name="display index",
        default=0,
        help_text="Index to control display order if not alphabetical, not required",
    )

    field_name = models.CharField(
        max_length=25,
        editable=False,
        default="",
        blank=True,
        help_text="Not required",
    )

    extra_value = models.CharField(max_length=250, default="")

    version = models.CharField(max_length=35, editable=False, default="1.0")

    objects = ListModelManager()

    def __str__(self) -> str:
        return self.display_name

    def save(self, *args, **kwargs):
        if not self.name:
            self.name = slugify(self.display_name).lower()
        if not self.plural_name:
            self.plural_name = f"{self.name}s"
        super().save(*args, **kwargs)

    def natural_key(self) -> tuple:
        return (self.name,)

    class Meta:
        abstract = True
        indexes = (models.Index(fields=["display_index", "display_name"]),)
        default_permissions = ("add", "change", "delete", "view", "export", "import")


class ListModelMixin(BaseListModelMixin):
    """Mixin for list data used in dropdown and radio widgets having
    display value and store value pairs.
    """

    id = models.AutoField(primary_key=True)

    class Meta(BaseListModelMixin.Meta):
        abstract = True
        indexes = BaseListModelMixin.Meta.indexes


class ListModelMixin2(BaseListModelMixin):
    """Mixin for list data used in dropdown and radio widgets having
    display value and store value pairs.

    Includes field "custom_name"
    """

    id = models.AutoField(primary_key=True)

    custom_name = models.CharField(
        max_length=25,
        default="",
        blank=True,
        help_text="A custom name/value to use on export instead of or in addition to `name`",
    )

    class Meta(BaseListModelMixin.Meta):
        abstract = True
        indexes = BaseListModelMixin.Meta.indexes


class ListUuidModelMixin(BaseListModelMixin, BaseUuidModel):
    """Mixin with UUID pk for list data used in dropdown
    and radio widgets having display value and store value pairs.
    """

    class Meta(BaseListModelMixin.Meta, BaseUuidModel.Meta):
        abstract = True
        indexes = (*BaseListModelMixin.Meta.indexes, *BaseUuidModel.Meta.indexes)



from django.apps import apps as django_apps
from django.core.management.base import BaseCommand, CommandError
from django.core.management.color import color_style

style = color_style()


class Command(BaseCommand):
    help = "Update search slugs if 'get_slugs' method definitions have changed."

    def add_arguments(self, parser):
        parser.add_argument(
            "model",
            nargs="+",
            type=str,
            help=(
                "a model name or list of model names in label_lower format. "
                "If no model names are specified then all models configured "
                "to use a search slug will be updated."
            ),
        )

    def handle(self, *args, **options):
        if options.get("model"):
            models = []
            for model_name in options.get("model"):
                models.append(django_apps.get_model(*model_name.split(".")))
        else:
            models = args or django_apps.get_models()
        for model in models:
            if "historical" not in model._meta.label_lower:
                if hasattr(model, "update_search_slugs"):
                    self.stdout.write(
                        self.style.WARNING(f"Updating '{model._meta.label_lower}' ..."),
                        ending="\r",
                    )
                    try:
                        model.objects.update_search_slugs()
                    except AttributeError as e:
                        if "update_search_slugs" in str(e):
                            raise CommandError(
                                "Missing manager method 'update_search_slugs'. "
                                f"See model {model._meta.label_lower}. Got {e!s}"
                            )
                        raise CommandError(e)
                    except Exception as e:
                        raise CommandError(
                            f"An exception occurred when updating model {model._meta.label_lower}. "
                            f"Got {e}"
                        )
                    else:
                        self.stdout.write(
                            self.style.SUCCESS(
                                f"Updating '{model._meta.label_lower}' ... Done"
                            )
                        )
                else:
                    self.stdout.write(f"------- {model._meta.label_lower}")



# Generated by Django 5.1 on 2024-08-14 18:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_qareports", "0013_rename_qareportnote_note_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="note",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Note",
                "verbose_name_plural": "Notes",
            },
        ),
        migrations.AlterField(
            model_name="note",
            name="status",
            field=models.CharField(
                choices=[
                    ("New", "New"),
                    ("PENDING", "Data is pending"),
                    ("closed", "Data not available"),
                ],
                default="New",
                max_length=25,
            ),
        ),
    ]



# Generated by Django 5.2.10 on 2026-01-24 20:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_pharmacy",
            "0115_rename_container_qty_historicalorderitem_container_unit_qty_and_more",
        ),
    ]

    operations = [
        migrations.RenameField(
            model_name="historicalorderitem",
            old_name="qty",
            new_name="item_qty_ordered",
        ),
        migrations.RenameField(
            model_name="historicalorderitem",
            old_name="unit_qty",
            new_name="unit_qty_pending",
        ),
        migrations.RenameField(
            model_name="historicalreceiveitem",
            old_name="qty",
            new_name="item_qty_received",
        ),
        migrations.RenameField(
            model_name="historicalreceiveitem",
            old_name="unit_qty",
            new_name="unit_qty_received",
        ),
        migrations.RenameField(
            model_name="orderitem",
            old_name="qty",
            new_name="item_qty_ordered",
        ),
        migrations.RenameField(
            model_name="orderitem",
            old_name="unit_qty",
            new_name="unit_qty_pending",
        ),
        migrations.RenameField(
            model_name="receiveitem",
            old_name="qty",
            new_name="item_qty_received",
        ),
        migrations.RenameField(
            model_name="receiveitem",
            old_name="unit_qty",
            new_name="unit_qty_received",
        ),
    ]



from django.contrib import admin
from django_audit_fields import audit_fieldset_tuple

from edc_model import DEFAULT_BASE_FIELDS

from .admin_site import edc_identifier_admin
from .models import IdentifierModel


@admin.register(IdentifierModel, site=edc_identifier_admin)
class IdentifierModelAdmin(admin.ModelAdmin):
    fieldsets = (
        [
            None,
            {
                "fields": (
                    "identifier",
                    "protocol_number",
                    "name",
                    "subject_identifier",
                    "site",
                    "model",
                    "sequence_number",
                    "identifier_type",
                    "linked_identifier",
                    "device_id",
                    "identifier_prefix",
                )
            },
        ],
        audit_fieldset_tuple,
    )

    list_display = (
        "identifier",
        "subject_identifier",
        "identifier_type",
        "site",
        "linked_identifier",
        "created",
        "user_created",
        "hostname_created",
    )
    list_filter = (
        "identifier_type",
        "name",
        "site",
        "device_id",
        "created",
        "user_created",
    )
    search_fields = ("identifier", "subject_identifier", "linked_identifier")

    def get_readonly_fields(self, request, obj=None) -> tuple[str, ...]:  # noqa: ARG002
        return (
            "identifier",
            "protocol_number",
            "subject_identifier",
            "name",
            "site",
            "model",
            "sequence_number",
            "identifier_type",
            "linked_identifier",
            "device_id",
            "identifier_prefix",
            *DEFAULT_BASE_FIELDS,
        )



from clinicedc_constants import NOT_APPLICABLE, PATIENT
from django.utils.translation import gettext_lazy as _

from .constants import (
    CANCELLED_APPT,
    COMPLETE_APPT,
    EXTENDED_APPT,
    IN_PROGRESS_APPT,
    INCOMPLETE_APPT,
    MISSED_APPT,
    NEW_APPT,
    ONTIME_APPT,
    SCHEDULED_APPT,
    SKIPPED_APPT,
    UNSCHEDULED_APPT,
)

DEFAULT_APPT_REASON_CHOICES = (
    (SCHEDULED_APPT, _("Scheduled (study-defined)")),
    (UNSCHEDULED_APPT, _("Unscheduled / Routine")),
)

APPT_STATUS = (
    (NEW_APPT, _("Not started")),
    (IN_PROGRESS_APPT, _("In Progress")),
    (INCOMPLETE_APPT, _("Incomplete")),
    (COMPLETE_APPT, _("Done")),
    (CANCELLED_APPT, _("Cancelled")),
    (SKIPPED_APPT, _("Skipped as per protocol")),
)

APPT_TIMING = (
    (ONTIME_APPT, _("On time (within window period)")),
    (MISSED_APPT, _("Missed")),
    (EXTENDED_APPT, _("Extended (extended window period for final appointment)")),
    (NOT_APPLICABLE, _("Not applicable")),
)

INFO_PROVIDER = (
    ("subject", _("Subject")),
    ("other", _("Other person")),
)

APPT_DATE_INFO_SOURCES = (
    ("health_records", _("Health record")),
    (PATIENT, _("Patient")),
    ("estimated", _("I estimated the date")),
)



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from dateutil.relativedelta import relativedelta
from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone

from ..exceptions import NotEvaluated
from ..formula import Formula
from .get_default_reportable_grades import get_default_reportable_grades
from .get_grade_for_value import get_grade_for_value
from .grading_data_model_cls import grading_data_model_cls
from .grading_exception_model_cls import grading_exception_model_cls

if TYPE_CHECKING:
    from ..models import ReferenceRangeCollection


__all__ = ["update_grading_data"]


def update_grading_data(
    reference_range_collection: ReferenceRangeCollection,
    grading_data: dict[str, list[Formula]] | None = None,
    reportable_grades: list[str] | None = None,
    reportable_grades_exceptions: dict[str, list[str]] | None = None,
    keep_existing: bool | None = None,
    create_missing_normal: bool | None = None,
):
    if not keep_existing:
        grading_data_model_cls().objects.filter(
            reference_range_collection=reference_range_collection
        ).delete()
    for label, formulas in grading_data.items():
        for formula in formulas:
            if get_reportable_grades(reference_range_collection, label, reportable_grades):
                formula_opts = {k: v for k, v in formula.__dict__.items() if k != "gender"}
                age_opts = {k: v for k, v in formula_opts.items() if "age" in k}
                for gender in formula.__dict__.get("gender"):
                    grading_data_model_cls().objects.create(
                        reference_range_collection=reference_range_collection,
                        label=label,
                        description=formula.description,
                        gender=gender,
                        **formula_opts,
                    )
                    for value in [formula.lower, formula.upper]:
                        if value:
                            try:
                                get_grade_for_value(
                                    reference_range_collection=reference_range_collection,
                                    label=label,
                                    value=value,
                                    units=formula_opts.get("units"),
                                    gender=gender,
                                    dob=timezone.now()
                                    - relativedelta(
                                        **{
                                            age_opts.get("age_units"): age_opts.get(
                                                "age_lower"
                                            )
                                        }
                                    ),
                                    report_datetime=timezone.now(),
                                    age_units=age_opts.get("age_units"),
                                    create_missing_normal=create_missing_normal,
                                )
                            except NotEvaluated as e:
                                sys.stdout.write(f"{e}\n")


def get_reportable_grades(
    reference_range_collection: ReferenceRangeCollection, label, reportable_grades
) -> list[int]:
    try:
        grading_exception = grading_exception_model_cls().objects.get(
            reference_range_collection=reference_range_collection, label=label
        )
    except ObjectDoesNotExist:
        reportable_grades = reportable_grades or get_default_reportable_grades()
    else:
        reportable_grades = grading_exception.grades.split(",")
    return reportable_grades



from clinicedc_constants import NOT_APPLICABLE
from django.db import models

from edc_constants.choices import YES_NO_NA
from edc_model.utils import duration_hm_to_timedelta
from edc_model.validators import hm_validator


def fasting_model_mixin_factory(
    prefix: str | None = None, verbose_names: dict | None = None, **kwargs
):
    prefix = "" if prefix is None else f"{prefix}_"
    verbose_names = verbose_names or {}

    class AbstractModel(models.Model):
        def save(self, *args, **kwargs):
            if duration_str := getattr(self, f"{prefix}fasting_duration_str", None):
                tdelta = duration_hm_to_timedelta(duration_str)
                setattr(self, f"{prefix}fasting_duration_delta", tdelta)
            super().save(*args, **kwargs)

        class Meta:
            abstract = True

    opts = {
        f"{prefix}fasting": models.CharField(
            verbose_name=verbose_names.get(
                f"{prefix}fasting",
                "Has the participant fasted?",
            ),
            max_length=15,
            choices=YES_NO_NA,
            default=NOT_APPLICABLE,
            blank=False,
            help_text="As reported by patient",
        ),
        f"{prefix}fasting_duration_str": models.CharField(
            verbose_name=verbose_names.get(
                f"{prefix}fasting_duration_str",
                "How long have they fasted in hours and/or minutes?",
            ),
            max_length=8,
            validators=[hm_validator],
            null=True,
            blank=True,
            help_text=(
                "As reported by patient. Duration of fast. Format is `HHhMMm`. "
                "For example 1h23m, 12h7m, etc"
            ),
        ),
        f"{prefix}fasting_duration_delta": models.DurationField(
            null=True,
            blank=True,
            help_text="system calculated to microseconds. (hours=microseconds/3.6e+9)",
        ),
    }

    opts.update(**kwargs)

    for name, fld_cls in opts.items():
        AbstractModel.add_to_class(name, fld_cls)

    return AbstractModel



from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from edc_model.admin import HistoricalModelAdminMixin
from edc_model_admin.history import SimpleHistoryAdmin

from ....admin_site import edc_pharmacy_history_admin
from ....models import Allocation
from ...model_admin_mixin import ModelAdminMixin


@admin.register(Allocation.history.model, site=edc_pharmacy_history_admin)
class AllocationHistoryAdmin(ModelAdminMixin, HistoricalModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Allocations (History)"
    change_form_title = "Pharmacy: Allocation (History)"
    change_list_note_url_name = "edc_pharmacy_admin:edc_pharmacy_allocation_changelist"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    list_display = (
        "subject_identifier",
        "history_type",
        "formatted_history_date",
        "code",
        "stock_changelist",
        "allocation_identifier",
        "revision",
        "history_id",
    )

    search_fields = ("code", "subject_identifier", "allocation_identifier")

    @admin.display(description="Stock #")
    def stock_changelist(self, obj):
        url = reverse("edc_pharmacy_history_admin:edc_pharmacy_historicalstock_changelist")
        url = f"{url}?q={obj.code}"
        context = dict(url=url, label=f"{obj.code}", title="Go to stock")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)



# Generated by Django 6.0 on 2025-03-15 00:12

import django.db.models.deletion
from django.db import migrations, models

import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0067_auto_20250314_1547"),
        ("sites", "0002_alter_domain_unique"),
    ]

    operations = [
        migrations.CreateModel(
            name="StockOut",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("report_model", models.CharField(max_length=50)),
                ("created", models.DateTimeField(default=django.utils.timezone.now)),
                ("subject_identifier", models.CharField(max_length=50, null=True)),
                (
                    "visit_code",
                    models.DecimalField(decimal_places=1, max_digits=8, null=True),
                ),
                ("appt_date", models.DateField(null=True)),
                ("baseline_date", models.DateField(null=True)),
                ("relative_days", models.IntegerField(null=True)),
                (
                    "site",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.DO_NOTHING, to="sites.site"
                    ),
                ),
            ],
            options={
                "verbose_name": "Stock out",
                "verbose_name_plural": "Stock out",
                "abstract": False,
                "default_permissions": ("view", "export", "viewallsites"),
                "indexes": [
                    models.Index(
                        fields=["subject_identifier", "site"],
                        name="edc_pharmac_subject_63b1ce_idx",
                    )
                ],
            },
        ),
    ]



from clinicedc_constants import CLOSED, NO

from edc_form_validators import FormValidator


class AeTmgFormValidator(FormValidator):
    def clean(self):
        self.required_if(
            NO, field="original_report_agreed", field_required="investigator_narrative"
        )
        self.applicable_if(
            NO,
            field="original_report_agreed",
            field_applicable="investigator_ae_classification",
        )
        self.validate_other_specify(field="investigator_ae_classification")
        self.required_if(
            CLOSED, field="report_status", field_required="report_closed_datetime"
        )



from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import UniqueConstraint
from django_crypto_fields.fields import EncryptedCharField

from edc_model.models import HistoricalRecords
from edc_sites.managers import CurrentSiteManager

from .exceptions import RandomizationListModelError
from .randomizer import RandomizationError
from .site_randomizers import site_randomizers

__all__ = ["RandomizationListManager", "RandomizationListModelMixin"]


class RandomizationListManager(models.Manager):
    def get_by_natural_key(self, sid):
        return self.get(sid=sid)


class RandomizationListModelMixin(models.Model):
    """
    A model mixin for the randomization list.

    The default expects and ACTIVE vs PLACEBO randomization. If
    yours differs, you need to re-declare field "assignment"
    and model method "treatment_description". The default
    `Randomizer` class MAY also need to be customized.
    """

    assignment = EncryptedCharField()

    randomizer_name = models.CharField(max_length=50, default="default")

    subject_identifier = models.CharField(  # noqa: DJ001
        verbose_name="Subject Identifier", max_length=50, null=True, unique=True
    )

    sid = models.IntegerField(unique=True)

    site_name = models.CharField(max_length=100)

    allocation = EncryptedCharField(verbose_name="Original integer allocation", null=True)

    allocated = models.BooleanField(default=False)

    allocated_datetime = models.DateTimeField(null=True)

    allocated_user = models.CharField(max_length=50, default="")

    allocated_site = models.ForeignKey(
        Site, null=True, on_delete=models.PROTECT, related_name="+"
    )

    verified = models.BooleanField(default=False)

    verified_datetime = models.DateTimeField(null=True)

    verified_user = models.CharField(max_length=50, default="")

    objects = RandomizationListManager()

    history = HistoricalRecords(inherit=True)

    on_site = CurrentSiteManager("allocated_site")

    def __str__(self):
        return f"{self.site_name}.{self.sid} subject={self.subject_identifier}"

    def save(self, *args, **kwargs):
        self.randomizer_name = self.randomizer_cls.name
        try:
            self.assignment_description  # noqa: B018
        except RandomizationError as e:
            raise RandomizationListModelError(e) from e
        try:
            Site.objects.get(name=self.site_name)
        except ObjectDoesNotExist as e:
            site_names = [obj.name for obj in Site.objects.all()]
            raise RandomizationListModelError(
                f"Invalid site name. Got {self.site_name}. Expected one of {site_names}."
            ) from e
        super().save(*args, **kwargs)

    @property
    def short_label(self) -> str:
        return f"{self.assignment} SID:{self.site_name}.{self.sid}"

    @property
    def randomizer_cls(self):
        return site_randomizers.get(self.randomizer_name)

    # customize if approriate
    @property
    def assignment_description(self) -> str:
        """May be overridden."""
        if self.assignment not in self.randomizer_cls.assignment_map:
            raise RandomizationError(
                f"Invalid assignment. Expected one of "
                f"{list(self.randomizer_cls.assignment_map.keys())}. "
                f"Got `{self.assignment}`. See "
            )
        return self.assignment

    def natural_key(self):
        return (self.sid,)

    class Meta:
        abstract = True
        constraints = (
            UniqueConstraint(
                fields=["site_name", "sid"],
                name="%(app_label)s_%(class)s_site_sid_uniq",
            ),
        )
        permissions = (("display_assignment", "Can display assignment"),)



from __future__ import annotations

from typing import TYPE_CHECKING

from ..views import PdfIntermediateView

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest
    from django.db.models import QuerySet


def print_selected_to_pdf_action(modeladmin, request: WSGIRequest, queryset: QuerySet):
    return PdfIntermediateView(request=request, model_pks=[o.id for o in queryset]).get(
        request,
        app_label=modeladmin.model._meta.app_label,
        model_name=modeladmin.model._meta.model_name,
    )



import sys
from typing import Any
from warnings import warn

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import (
    MultipleObjectsReturned,
    ObjectDoesNotExist,
    ValidationError,
)
from django.core.management.color import color_style

from ..constants import PII, PII_VIEW
from ..utils import make_view_only_group_permissions

style = color_style()

INVALID_APP_LABEL = "invalid_app_label"


class PermissionsCodenameError(Exception):
    pass


class PermissionsCreatorError(ValidationError):
    pass


class CodenameDoesNotExist(Exception):  # noqa: N818
    pass


class GroupUpdater:
    default_model_name = "edcpermissions"

    def __init__(
        self,
        groups: dict[str, list[str]] | None = None,
        apps=None,
        verbose=None,
        pii_models: list | None = None,
        custom_permissions_tuples: dict | None = None,
        warn_only=None,
        codename_prefixes: list | None = None,
    ):
        self.apps = apps or django_apps
        self.content_type_model_cls = self.apps.get_model("contenttypes.contenttype")
        self.custom_permissions_tuples = custom_permissions_tuples
        self.group_model_cls = self.apps.get_model("auth.group")
        self.group_names = list(groups.keys())
        self.groups = groups
        self.permission_model_cls = self.apps.get_model("auth.permission")
        self.pii_models = pii_models or []
        self.verbose = verbose
        self.warn_only = getattr(settings, "EDC_AUTH_CODENAMES_WARN_ONLY", warn_only)

        # TODO: perhaps change edc_navbar codenames to have a prefix such as `access`
        self.codename_prefixes = codename_prefixes or [
            "view",
            "add",
            "change",
            "delete",
            "import",
            "export",
            "nav",
            "display",
            "special",
            "viewallsites",
        ]

    def update_groups(self):
        if self.verbose:
            sys.stdout.write(style.MIGRATE_HEADING(" - Updating groups:\n"))
        for group_name, codenames in self.groups.items():
            self.update_group(group_name, codenames, create_group=True)
        for group in self.group_model_cls.objects.exclude(name__in=[PII, PII_VIEW]):
            self.remove_pii_permissions_from_group(group)
        self.group_model_cls.objects.exclude(name__in=self.group_names).delete()
        if self.verbose:
            sys.stdout.write("   Done.\n")

    def update_group(self, group_name, codenames, create_group=None):
        if self.verbose:
            sys.stdout.write(f"   * {group_name.lower()}\n")
        try:
            group = self.group_model_cls.objects.get(name=group_name)
        except ObjectDoesNotExist as e:
            if not create_group:
                raise ObjectDoesNotExist(f"{e} Got {group_name}") from e
            group = self.group_model_cls.objects.create(name=group_name)
        else:
            group.permissions.clear()
        self.add_permissions_to_group_by_codenames(group, codenames)
        self.remove_historical_group_permissions(group)
        make_view_only_group_permissions("site", group)

    def add_permissions_to_group_by_codenames(self, group=None, codenames=None):
        if codenames:
            permissions = self.get_permissions_qs_from_codenames(codenames)
            for permission in permissions:
                group.permissions.add(permission)

    def get_permissions_qs_from_codenames(
        self, codenames: list[Any], allow_multiple_objects: bool | None = None
    ):
        """Returns a list of permission model instances for the given
        codenames.

        codenames:  a list of codenames or a list of functions that
                      returns a list. Combining codenames and funcs
                      in a list also works.
        """
        permissions = []
        expanded_codenames = []
        for item in codenames:
            try:
                expanded_codenames.extend(item())
            except TypeError:
                expanded_codenames.append(item)
        for dotted_codename in expanded_codenames:
            try:
                app_label, codename = self.get_from_dotted_codename(dotted_codename)
            except PermissionsCodenameError as e:
                warn(str(e), stacklevel=2)
            else:
                # if you add extra codenames you must also add custom
                # codename tuples to the Permissions model before you
                # get here. See also changes to default_permissions from
                # model Meta class BaseUuidModel.Meta.
                # See also site_auths.add_custom_permissions_tuples
                try:
                    permissions.append(
                        self.permission_model_cls.objects.get(
                            codename=codename, content_type__app_label=app_label
                        )
                    )
                except ObjectDoesNotExist as e:
                    errmsg = f"{e} Got codename={codename},app_label={app_label}"
                    if not self.warn_only:
                        raise CodenameDoesNotExist(errmsg) from e
                    warn(style.ERROR(errmsg), stacklevel=2)
                except MultipleObjectsReturned as e:
                    if not allow_multiple_objects:
                        self.delete_and_raise_on_duplicate_codenames(
                            codename, app_label, exception=e
                        )
                    permissions.extend(
                        [
                            obj
                            for obj in self.permission_model_cls.objects.filter(
                                codename=codename, content_type__app_label=app_label
                            )
                        ]
                    )
        return permissions

    def get_from_dotted_codename(self, codename=None):
        """Returns a tuple of app_label, codename.

        Validates the codename format, '<app_label>.<some_codename>',
        and the `app_label` in a given codename.
        """
        if not codename:
            raise PermissionsCodenameError("Invalid codename. May not be None.")
        try:
            app_label, _codename = codename.split(".")
        except ValueError as e:
            raise PermissionsCodenameError(
                f"Invalid dotted codename. {e} Got {codename}."
            ) from e
        else:
            try:
                self.apps.get_app_config(app_label)
            except LookupError as e:
                raise PermissionsCodenameError(
                    "Invalid app_label in codename. Expected format "
                    f"'<app_label>.<some_codename>'. Got {codename}."
                ) from e
        prefix = _codename.split("_")[0]
        if prefix not in self.codename_prefixes:
            raise PermissionsCodenameError(
                f"Invalid codename prefix. Expected one of {self.codename_prefixes}. "
                f"Got {_codename}."
            )
        return app_label, _codename

    def remove_pii_permissions_from_group(self, group):
        for model in self.pii_models:
            self.remove_permissions_by_model(group, model)

    @staticmethod
    def remove_historical_group_permissions(group=None, model=None):
        """Removes permissions for historical models from this group.

        Default removes all except `view`.
        """
        make_view_only_group_permissions("_historical", group=group, model=model)

    def remove_permissions_by_model(self, group=None, model=None):
        try:
            model_cls = self.apps.get_model(model)
        except LookupError as e:
            warn(f"Unable to remove permissions. {e}. Got {model}")
        else:
            content_type = self.content_type_model_cls.objects.get_for_model(model_cls)
            for permission in self.permission_model_cls.objects.filter(
                content_type=content_type
            ):
                group.permissions.remove(permission)

    def create_custom_permissions_from_tuples(self):
        for model, codename_tuples in self.custom_permissions_tuples.items():
            self.create_permissions_from_tuples(model, codename_tuples)

    def create_permissions_from_tuples(self, model=None, codename_tuples=None):
        """Creates custom permissions on model `model`."""
        if codename_tuples:
            try:
                model_cls = self.apps.get_model(model)
            except LookupError as e:
                warn(f"{e}. Got {model}", stacklevel=2)
            else:
                content_type = self.content_type_model_cls.objects.get_for_model(model_cls)
                for codename_tpl in codename_tuples:
                    app_label, codename, name = self.get_from_codename_tuple(
                        codename_tpl, model_cls._meta.app_label
                    )
                    try:
                        self.permission_model_cls.objects.get(
                            codename=codename, content_type=content_type
                        )
                    except ObjectDoesNotExist:
                        self.permission_model_cls.objects.create(
                            name=name, codename=codename, content_type=content_type
                        )
                    self.verify_codename_exists(f"{app_label}.{codename}", content_type)

    def verify_codename_exists(self, codename, content_type):
        permission = None
        app_label, codename = self.get_from_dotted_codename(codename)
        try:
            permission = self.permission_model_cls.objects.get(
                codename=codename, content_type=content_type
            )
        except ObjectDoesNotExist as e:
            raise CodenameDoesNotExist(
                f"Unable to verify codename. {e} Got '{app_label}.{codename}'"
            ) from e
        except MultipleObjectsReturned as e:
            self.delete_and_raise_on_duplicate_codenames(codename, app_label, exception=e)
        return permission

    def delete_and_raise_on_duplicate_codenames(
        self, codename: str, app_label: str, exception=None
    ):
        self.permission_model_cls.objects.filter(
            content_type__app_label=app_label,
            content_type__model=self.default_model_name,
        ).delete()
        self.permission_model_cls.objects.filter(codename=codename).delete()
        raise CodenameDoesNotExist(
            f"Unable to verify codename. {exception or ''} Got '{app_label}.{codename}'. \n"
            "YOU NEED TO RUN MIGRATE AGAIN!\n"
        )

    @staticmethod
    def get_from_codename_tuple(codename_tpl, app_label=None):
        try:
            value, name = codename_tpl
        except ValueError as e:
            raise ValueError(f"{e} Got {codename_tpl}") from e
        _app_label, codename = value.split(".")
        if app_label and _app_label != app_label:
            raise PermissionsCreatorError(
                f"app_label in permission codename does not match. "
                f"Expected {app_label}. Got {_app_label}. "
                f"See {codename_tpl}.",
                code=INVALID_APP_LABEL,
            )
        return _app_label, codename, name

    def remove_permissions_by_codenames(
        self,
        group,
        codenames: list[str],
        allow_multiple_objects: bool | None = None,
    ):
        """Remove the given codenames from the given group."""
        permissions = self.get_permissions_qs_from_codenames(
            codenames, allow_multiple_objects=allow_multiple_objects
        )
        for permission in permissions:
            group.permissions.remove(permission)



from django.apps import apps as django_apps

from edc_auth.site_auths import site_auths
from edc_export.constants import EXPORT


def update_site_auths():
    if django_apps.is_installed("edc_export"):
        site_auths.update_group(
            "edc_metadata.export_crfmetadata",
            "edc_metadata.export_requisitionmetadata",
            name=EXPORT,
        )


update_site_auths()



from clinicedc_constants import INCOMPLETE
from django.apps import apps as django_apps

from .model_mixins import CrfStatusModelMixin
from .update_crf_status_for_instance import update_crf_status_for_instance


def update_crf_status_command(app_label: str | None = None) -> None:
    if app_label:
        app_configs = [django_apps.get_app_config(app_label)]
    else:
        app_configs = django_apps.get_app_configs()

    print("Updating CRF Status model for instances set to crf_status=incomplete")
    for app_config in app_configs:
        print(f"  * updating {app_config.name}")
        for model in app_config.get_models():
            if issubclass(model, (CrfStatusModelMixin,)):
                print(f"    - {model._meta.label_lower}")
                for obj in model.objects.filter(crf_status=INCOMPLETE):
                    update_crf_status_for_instance(obj)



# default group names

ACCOUNT_MANAGER = "ACCOUNT_MANAGER"
ADMINISTRATION = "ADMINISTRATION"
AUDITOR = "AUDITOR"
CELERY_MANAGER = "CELERY_MANAGER"
CLINIC = "CLINIC"
CLINIC_SUPER = "CLINIC_SUPER"
EVERYONE = "EVERYONE"
PII = "PII"
PII_VIEW = "PII_VIEW"

# default role names
ACCOUNT_MANAGER_ROLE = "account_manager"
AUDITOR_ROLE = "auditor"
CLINICIAN_ROLE = "clinician"
CLINICIAN_SUPER_ROLE = "clinician_super"
CUSTOM_ROLE = "custom"
NURSE_ROLE = "research_nurse"
STAFF_ROLE = "staff"
STATISTICIAN = "statistician"

# keys for the site_auths registry
UPDATE_ROLES_KEY = "update_roles"
UPDATE_GROUPS_KEY = "update_groups"
PRE_UPDATE_FUNCS_KEY = "pre_update_funcs"
POST_UPDATE_FUNCS_KEY = "post_update_funcs"
GROUPS_KEY = "groups"
ROLES_KEY = "roles"
PII_MODELS_KEY = "pii_models"
CUSTOM_PERMISSIONS_TUPLES_KEY = "custom_permissions_tuples"

TMG_ROLE = "tmg"
AE_ROLE = "ae_role"



from .death_report_extra_fields_model_mixin import DeathReportExtraFieldsModelMixin
from .death_report_model_mixin import DeathReportModelMixin
from .death_report_tmg_model_mixin import (
    DeathReportTmgManager,
    DeathReportTmgModelMixin,
    DeathReportTmgSecondManager,
    DeathReportTmgSecondSiteManager,
    DeathReportTmgSiteManager,
)
from .death_report_tmg_second_model_mixin import DeathReportTmgSecondModelMixin
from .simple_death_report_model_mixin import SimpleDeathReportModelMixin



# Generated by Django 5.2.6 on 2025-09-18 00:24

import django.utils.timezone
import edc_model.validators.date
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_lab", "0036_alter_aliquot_comment_alter_aliquot_device_created_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalorder",
            name="order_datetime",
            field=models.DateTimeField(
                default=django.utils.timezone.now,
                validators=[edc_model.validators.date.datetime_not_future],
            ),
        ),
        migrations.AlterField(
            model_name="order",
            name="order_datetime",
            field=models.DateTimeField(
                default=django.utils.timezone.now,
                validators=[edc_model.validators.date.datetime_not_future],
            ),
        ),
    ]



from datetime import datetime
from typing import Any
from zoneinfo import ZoneInfo

from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from edc_dashboard.url_names import url_names
from edc_form_validators import INVALID_ERROR


def get_appointment_url(appointment_pk: str, subject_identifier: str) -> str:
    rev_url = reverse(
        "edc_appointment_admin:edc_appointment_appointment_change",
        args=(appointment_pk,),
    )
    rev_url = (
        f"{rev_url}?next={url_names.get('subject_dashboard_url')},"
        f"subject_identifier"
        f"&subject_identifier={subject_identifier}"
    )
    return rev_url


def validate_appt_datetime_unique(
    form_validator: Any,
    appointment: Any,
    appt_datetime: datetime,
    form_field: str | None = None,
):
    """Assert one visit report per day"""
    if appt_datetime:
        form_field = form_field or "appt_datetime"
        appt_datetime_utc = appt_datetime.astimezone(ZoneInfo("UTC"))
        opts = {
            "subject_identifier": appointment.subject_identifier,
            "visit_schedule_name": appointment.visit_schedule_name,
            "schedule_name": appointment.schedule_name,
            "appt_datetime__date": appt_datetime_utc.date(),
        }
        other_appts = appointment.__class__.objects.filter(**opts).exclude(id=appointment.id)
        if other_appts.count() > 1:
            raise form_validator.raise_validation_error(
                {form_field: "An appointment already exists for this date (M)"},
                INVALID_ERROR,
            )
        if other_appts.count() == 1:
            if appointment and other_appts[0].id != appointment.id:
                appointment_url = get_appointment_url(
                    (
                        appointment.id
                        if appointment.visit_code_sequence > 0
                        else other_appts[0].id
                    ),
                    appointment.subject_identifier,
                )
                if other_appts[0].visit_code_sequence == 0:
                    msg = format_html(
                        (
                            "This appointment conflicts with the scheduled appointment "
                            '{visit_code}.0. Consider editing <A href="{appointment_url}"> '
                            "this appointment</A> first."
                        ),
                        visit_code=other_appts[0].visit_code,
                        appointment_url=mark_safe(appointment_url),  # nosec B703, B308
                    )
                else:
                    phrase = (
                        "A scheduled"
                        if other_appts[0].visit_code_sequence == 0
                        else "An unscheduled"
                    )
                    msg = format_html(
                        "{phrase} appointment already exists for this date. "
                        'See <A title="Edit appointment" href="{appointment_url}">'
                        "appointment {visit_code}. {visit_code_sequence}</A>",
                        phrase=phrase,
                        appointment_url=mark_safe(appointment_url),  # nosec B703, B308
                        visit_code=other_appts[0].visit_code,
                        visit_code_sequence=other_appts[0].visit_code_sequence,
                    )
                raise form_validator.raise_validation_error(
                    {form_field: msg},
                    INVALID_ERROR,
                )



# Generated by Django 5.2.6 on 2025-09-22 12:31

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_auth", "0035_alter_edcpermissions_device_created_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="userprofile",
            name="alternate_email",
            field=models.EmailField(
                blank=True,
                default="",
                max_length=254,
                verbose_name="Alternate email address",
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="clinic_label_printer",
            field=models.CharField(
                blank=True,
                default="",
                help_text='Change in <a href="/edc_label/">Edc Label Administration</a>',
                max_length=100,
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="export_format",
            field=models.CharField(
                blank=True,
                choices=[
                    ("CSV", "CSV (delimited by pipe `|`)"),
                    (114, "Stata v10 or later"),
                    (117, "Stata v13 or later"),
                    (118, "Stata v14 or later"),
                    (119, "Stata v15 or later"),
                ],
                default="CSV",
                help_text="Note: requires export permissions",
                max_length=25,
                verbose_name="Export format",
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="job_title",
            field=models.CharField(blank=True, default="", max_length=100),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="lab_label_printer",
            field=models.CharField(
                blank=True,
                default="",
                help_text='Change in <a href="/edc_label/">Edc Label Administration</a>',
                max_length=100,
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="mobile",
            field=models.CharField(
                blank=True,
                default="",
                help_text="e.g. +1234567890",
                max_length=25,
                validators=[django.core.validators.RegexValidator(regex="^\\+\\d+")],
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="print_server",
            field=models.CharField(
                blank=True,
                default="",
                help_text='Change in <a href="/edc_label/">Edc Label Administration</a>',
                max_length=100,
            ),
        ),
    ]



from django.core.management.base import BaseCommand, CommandError
from edc_registration.models import RegisteredSubject
from tqdm import tqdm

from edc_appointment.utils import refresh_appointments


class Command(BaseCommand):
    help = (
        "Validate appointment visit code sequences relative to appt_datetime "
        "and reset if needed."
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "--visit-schedule-name",
            dest="visit_schedule_name",
            default=None,
            help="Visit schedule name",
        )

        parser.add_argument(
            "--schedule-name",
            dest="schedule_name",
            default=None,
            help="Schedule name",
        )

    def handle(self, *args, **options):
        visit_schedule_name = options.get("visit_schedule_name")
        if not visit_schedule_name:
            raise CommandError("--visit-schedule-name is required")
        schedule_name = options.get("schedule_name")
        if not schedule_name:
            raise CommandError("--schedule-name is required")
        qs = RegisteredSubject.objects.all()
        for obj in tqdm(qs, total=qs.count()):
            refresh_appointments(
                obj.subject_identifier,
                visit_schedule_name,
                schedule_name,
                warn_only=True,
                skip_get_current_site=True,
            )



from copy import copy
from typing import Any

from django.urls.base import reverse

from edc_dashboard.url_names import url_names

from .base_box_item_listboard_view import BaseBoxItemListboardView


class ManageBoxListboardView(BaseBoxItemListboardView):
    action_name = "manage"
    form_action_url = "manage_box_item_form_action_url"  # url_name
    listboard_url = "manage_box_listboard_url"  # url_name
    listboard_template = "manage_box_listboard_template"
    verify_box_listboard_url = "verify_box_listboard_url"  # url_name
    listboard_model = "edc_lab.boxitem"
    navbar_selected_item = "pack"
    search_form_url = "manage_box_listboard_url"  # url_name

    @property
    def url_kwargs(self):
        return {"action_name": self.action_name, "box_identifier": self.box_identifier}

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        url_kwargs = copy(self.url_kwargs)
        url_kwargs["position"] = 1
        url_kwargs["action_name"] = "verify"
        kwargs.update(
            verify_box_listboard_url_reversed=reverse(
                url_names.get(self.verify_box_listboard_url), kwargs=url_kwargs
            )
        )
        return super().get_context_data(**kwargs)



from django import forms

from ...constants import CENTRAL_LOCATION
from ...models import StockTransfer


class StockTransferForm(forms.ModelForm):
    @property
    def to_location(self):
        return self.cleaned_data.get("to_location") or self.instance.to_location

    @property
    def from_location(self):
        return self.cleaned_data.get("from_location") or self.instance.from_location

    def clean(self):
        cleaned_data = super().clean()

        # assuming all fields expect 'comment' are set to readonly on EDIT
        if not self.instance.id:
            # items_qs = StockTransferItem.objects.filter(stock_transfer__pk=self.instance.pk)
            # if items_qs.count() == 0:
            #     raise forms.ValidationError("Nothing to transfer")

            # TO/FROM locations cannot be the same
            if self.to_location == self.from_location:
                raise forms.ValidationError(
                    {
                        "__all__": "Invalid location combination. 'TO' and 'FROM' locations "
                        "cannot be the same"
                    }
                )

            # at least one location must be CENTRAL
            if CENTRAL_LOCATION not in [self.to_location.name, self.from_location.name]:
                raise forms.ValidationError(
                    {"__all__": "Invalid location combination. One location must be Central"}
                )

            # check the current stock site if not the same as to_location
            # if items_qs[0].stock.location == self.to_location:
            #     raise forms.ValidationError(
            #         {
            #             "__all__": "Invalid 'TO' location. Stock is already at this location. "
            #             f"Got {self.to_location}"
            #         }
            #     )

            # if FROM central, TO must be subject's site
            # if self.from_location.name == CENTRAL_LOCATION and (
            #     items_qs[0].stock.allocation.registered_subject.site != self.to_location.site
            # ):
            #     raise forms.ValidationError(
            #         {
            #             "__all__": (
            #                 "Invalid 'TO' location. Does not match the allocated subject's "
            #                 "location / study site."
            #             )
            #         }
            #     )

            # if TO central, FROM must be subject's site
            # if self.to_location.name == CENTRAL_LOCATION and (
            #     items_qs[0].stock.allocation.registered_subject.site != self.from_location.site
            # ):
            #     raise forms.ValidationError(
            #         {
            #             "__all__": (
            #                 "Invalid 'FROM' location. Does not match the allocated subject's "
            #                 "location / study site."
            #             )
            #         }
            #     )

            # if cleaned_data.get(
            #     "item_count"
            # ) is not None and items_qs.count() > cleaned_data.get("item_count"):
            #     raise forms.ValidationError(
            #         {
            #             "__all__": (
            #                 "Invalid item count. Expected a value "
            #                 f"greater than or equal to {items_qs.count()}"
            #             )
            #         }
            #     )

        return cleaned_data

    class Meta:
        model = StockTransfer
        fields = "__all__"
        help_text = {"transfer_identifier": "(read-only)"}  # noqa: RUF012
        widgets = {  # noqa: RUF012
            "transfer_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
        }



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple
from edc_fieldsets.fieldset import Fieldset

from ..admin_site import edc_lab_admin
from ..forms import AliquotForm
from ..models import Aliquot
from .base_model_admin import BaseModelAdmin

aliquot_identifiers_fields = (
    "subject_identifier",
    "requisition_identifier",
    "parent_identifier",
    "identifier_prefix",
)

aliquot_identifiers_fieldset_tuple = Fieldset(
    *aliquot_identifiers_fields, section="Identifiers"
).fieldset


@admin.register(Aliquot, site=edc_lab_admin)
class AliquotAdmin(BaseModelAdmin, admin.ModelAdmin):
    form = AliquotForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "aliquot_identifier",
                    "aliquot_datetime",
                    "aliquot_type",
                    "numeric_code",
                    "alpha_code",
                    "condition",
                )
            },
        ),
        aliquot_identifiers_fieldset_tuple,
        ("Shipping", {"classes": ("collapse",), "fields": ("shipped",)}),
        audit_fieldset_tuple,
    )

    search_fields = ("aliquot_identifier", "subject_identifier")

    radio_fields = {"condition": admin.VERTICAL}  # noqa: RUF012

    def get_list_display(self, request) -> tuple:
        list_display = super().get_list_display(request)
        custom_fields = (
            "aliquot_identifier",
            "subject_identifier",
            "aliquot_datetime",
            "aliquot_type",
            "condition",
        )
        return tuple(set(custom_fields + list_display))

    def get_list_filter(self, request) -> tuple:
        list_filter = super().get_list_filter(request)
        custom_fields = ("aliquot_datetime", "aliquot_type", "condition")
        return tuple(set(custom_fields + list_filter))

    def get_readonly_fields(self, request, obj=None) -> tuple:
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        return tuple(set(readonly_fields + aliquot_identifiers_fields))


class AliquotInlineAdmin(admin.TabularInline):
    model = Aliquot
    extra = 0
    fields = ("aliquot_identifier",)



# Generated by Django 6.0 on 2026-01-14 21:20

from django.db import migrations
from django.db.migrations import RunPython
from tqdm import tqdm

from edc_pharmacy.models import StockTransferItem, StorageBinItem


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0096_auto_20260114_1800"),
    ]

    operations = []
